@@ -18,7 +18,7 @@
 #define PARTITION_TABLE_SIZE_GPT   34
 
 typedef struct _PuEmmcInput {
-    gchar *uri;
+    gchar *filename;
     gchar *md5sum;
     gchar *sha256sum;
 
@@ -30,6 +30,7 @@ typedef struct _PuEmmcPartition {
     gchar *partuuid;
     PedPartitionType type;
     gchar *filesystem;
+    gchar *mkfs_extra_args;
     PedSector size;
     PedSector offset;
     PedSector block_size;
@@ -43,11 +44,15 @@ typedef struct _PuEmmcBinary {
     PuEmmcInput *input;
 } PuEmmcBinary;
 typedef struct _PuEmmcBootPartitions {
-    gboolean enable;
-    PedSector input_offset;
-    PedSector output_offset;
-    PuEmmcInput *input;
+    guint enable;
+    gboolean boot_ack;
+    GList *input;
 } PuEmmcBootPartitions;
+typedef struct _PuEmmcControls {
+    gchar *hwreset;
+    gchar *bootbus;
+    PuEmmcBootPartitions *boot_partitions;
+} PuEmmcControls;
 typedef struct _PuEmmcClean {
     PedSector size;
     PedSector offset;
@@ -67,7 +72,7 @@ struct _PuEmmc {
     GList *partitions;
     GList *clean;
     GList *raw;
-    PuEmmcBootPartitions *emmc_boot_partitions;
+    PuEmmcControls *mmc_controls;
 };
 
 G_DEFINE_TYPE(PuEmmc, pu_emmc, PU_TYPE_FLASH)
@@ -140,7 +145,9 @@ emmc_create_partition(PuEmmc *self,
         return FALSE;
     }
 
-    if (ped_disk_type_check_feature(part->disk->type, PED_DISK_TYPE_PARTITION_NAME)) {
+    /* Set PARTLABEL on GPT disks */
+    if (ped_disk_type_check_feature(part->disk->type, PED_DISK_TYPE_PARTITION_NAME) &&
+        partition->label) {
         ped_partition_set_name(part, partition->label);
     }
 
@@ -286,7 +293,8 @@ pu_emmc_write_data(PuFlash *flash,
 
         g_debug("Creating filesystem '%s' on '%s'", part->filesystem, part_path);
 
-        if (!pu_make_filesystem(part_path, part->filesystem, error))
+        if (!pu_make_filesystem(part_path, part->filesystem, part->label,
+                                part->mkfs_extra_args, error))
             return FALSE;
 
         if (!part->input) {
@@ -304,9 +312,9 @@ pu_emmc_write_data(PuFlash *flash,
             PuEmmcInput *input = i->data;
             g_autofree gchar *path = NULL;
 
-            path = pu_path_from_uri(input->uri, prefix, error);
+            path = pu_path_from_filename(input->filename, prefix, error);
             if (path == NULL) {
-                g_prefix_error(error, "Failed parsing input URI for partition: ");
+                g_prefix_error(error, "Failed parsing input filename for partition: ");
                 return FALSE;
             }
 
@@ -335,6 +343,8 @@ pu_emmc_write_data(PuFlash *flash,
                     return FALSE;
                 if (!pu_resize_filesystem(part_path, error))
                     return FALSE;
+                if (!pu_set_ext_label(part_path, part->label, error))
+                    return FALSE;
             } else {
                 if (!pu_mount(part_path, part_mount, NULL, NULL, error))
                     return FALSE;
@@ -369,9 +379,9 @@ pu_emmc_write_data(PuFlash *flash,
         PuEmmcInput *input = bin->input;
         g_autofree gchar *path = NULL;
 
-        path = pu_path_from_uri(input->uri, prefix, error);
+        path = pu_path_from_filename(input->filename, prefix, error);
         if (path == NULL) {
-            g_prefix_error(error, "Failed parsing input URI for binary: ");
+            g_prefix_error(error, "Failed parsing input filename for binary: ");
             return FALSE;
         }
 
@@ -393,62 +403,81 @@ pu_emmc_write_data(PuFlash *flash,
                 return FALSE;
         }
 
-        g_debug("Writing raw data: uri=%s input_offset=%lld output_offset=%lld",
-                input->uri, bin->input_offset, bin->output_offset);
+        g_debug("Writing raw data: filename=%s input_offset=%lld output_offset=%lld",
+                input->filename, bin->input_offset, bin->output_offset);
 
         if (!pu_write_raw(path, self->device->path, self->device,
                           bin->input_offset, bin->output_offset, 0, error))
             return FALSE;
     }
 
-    if (self->emmc_boot_partitions) {
-        PuEmmcInput *input = self->emmc_boot_partitions->input;
-        g_autofree gchar *path = NULL;
+    if (self->mmc_controls) {
+        PuEmmcBootPartitions *boot_partitions = NULL;
 
-        path = pu_path_from_uri(input->uri, prefix, error);
-        if (path == NULL) {
-            g_prefix_error(error, "Failed parsing input URI for eMMC boot partition: ");
-            return FALSE;
-        }
-
-        if (g_str_equal(path, "")) {
-            g_set_error(error, PU_ERROR, PU_ERROR_FLASH_DATA,
-                        "No input specified for eMMC boot partition");
-            return FALSE;
-        }
-
-        if (!g_str_equal(input->md5sum, "") && !skip_checksums) {
-            g_debug("Checking MD5 sum of input file '%s'", path);
-            if (!pu_checksum_verify_file(path, input->md5sum,
-                                         G_CHECKSUM_MD5, error))
+        if (pu_has_bootpart(self->device->path)) {
+            if (!pu_set_hwreset(self->device->path,
+                                   self->mmc_controls->hwreset, error))
                 return FALSE;
-        }
-        if (!g_str_equal(input->sha256sum, "") && !skip_checksums) {
-            g_debug("Checking SHA256 sum of input file '%s'", path);
-            if (!pu_checksum_verify_file(path, input->sha256sum,
-                                         G_CHECKSUM_SHA256, error))
+
+            if (!pu_set_bootbus(self->device->path,
+                                self->mmc_controls->bootbus, error))
                 return FALSE;
         }
 
-        g_debug("Writing eMMC boot partitions: uri=%s input_offset=%lld output_offset=%lld",
-                input->uri, self->emmc_boot_partitions->input_offset,
-                self->emmc_boot_partitions->output_offset);
+        boot_partitions = self->mmc_controls->boot_partitions;
+        if (boot_partitions && pu_has_bootpart(self->device->path)) {
+            GList *input = boot_partitions->input;
 
-        if (!pu_write_raw_bootpart(path, self->device, 0,
-                                   self->emmc_boot_partitions->input_offset,
-                                   self->emmc_boot_partitions->output_offset,
-                                   error))
-            return FALSE;
+            if (!pu_bootpart_enable(self->device->path, boot_partitions->enable,
+                                    boot_partitions->boot_ack, error))
+                return FALSE;
 
-        if (!pu_write_raw_bootpart(path, self->device, 1,
-                                   self->emmc_boot_partitions->input_offset,
-                                   self->emmc_boot_partitions->output_offset,
-                                   error))
-            return FALSE;
+            for (GList *i = input; i != NULL; i = i->next) {
+                PuEmmcBinary *bin = i->data;
+                g_autofree gchar *path = NULL;
 
-        if (!pu_bootpart_enable(self->device->path,
-                                self->emmc_boot_partitions->enable, error))
-            return FALSE;
+                path = pu_path_from_filename(bin->input->filename, prefix, error);
+                if (path == NULL) {
+                    g_prefix_error(error, "Failed parsing input filename for eMMC boot partition: ");
+                    return FALSE;
+                }
+
+                if (g_str_equal(path, "")) {
+                    g_set_error(error, PU_ERROR, PU_ERROR_FLASH_DATA,
+                                "No input specified for eMMC boot partition");
+                    return FALSE;
+                }
+
+                if (!g_str_equal(bin->input->md5sum, "") && !skip_checksums) {
+                    g_debug("Checking MD5 sum of input file '%s'", path);
+                    if (!pu_checksum_verify_file(path, bin->input->md5sum,
+                                                 G_CHECKSUM_MD5, error))
+                        return FALSE;
+                }
+                if (!g_str_equal(bin->input->sha256sum, "") && !skip_checksums) {
+                    g_debug("Checking SHA256 sum of input file '%s'", path);
+                    if (!pu_checksum_verify_file(path, bin->input->sha256sum,
+                                                 G_CHECKSUM_SHA256, error))
+                        return FALSE;
+                }
+
+                g_debug("Writing eMMC boot partitions: filename=%s input_offset=%lld output_offset=%lld",
+                        bin->input->filename, bin->input_offset,
+                        bin->output_offset);
+
+                if (!pu_write_raw_bootpart(path, self->device, 0,
+                                           bin->input_offset,
+                                           bin->output_offset,
+                                           error))
+                    return FALSE;
+
+                if (!pu_write_raw_bootpart(path, self->device, 1,
+                                           bin->input_offset,
+                                           bin->output_offset,
+                                           error))
+                    return FALSE;
+            }
+        }
     }
 
     g_debug("%s: Finished", G_STRFUNC);
@@ -466,10 +495,11 @@ pu_emmc_class_finalize(GObject *object)
         g_free(part->label);
         g_free(part->partuuid);
         g_free(part->filesystem);
+        g_free(part->mkfs_extra_args);
         g_list_free(g_steal_pointer(&part->flags));
         for (GList *i = part->input; i != NULL; i = i->next) {
             PuEmmcInput *in = i->data;
-            g_free(in->uri);
+            g_free(in->filename);
             g_free(in->md5sum);
             g_free(in->sha256sum);
             g_free(in);
@@ -482,20 +512,31 @@ pu_emmc_class_finalize(GObject *object)
 
     for (GList *b = emmc->raw; b != NULL; b = b->next) {
         PuEmmcBinary *bin = b->data;
-        g_free(bin->input->uri);
+        g_free(bin->input->filename);
         g_free(bin->input->md5sum);
         g_free(bin->input->sha256sum);
         g_free(bin->input);
+        g_free(bin);
     }
     g_list_free(g_steal_pointer(&emmc->raw));
 
-    if (emmc->emmc_boot_partitions) {
-        PuEmmcInput *input = emmc->emmc_boot_partitions->input;
-        g_free(input->uri);
-        g_free(input->md5sum);
-        g_free(input->sha256sum);
-        g_free(input);
-        g_free(emmc->emmc_boot_partitions);
+    if (emmc->mmc_controls) {
+        g_free(emmc->mmc_controls->hwreset);
+        g_free(emmc->mmc_controls->bootbus);
+        if (emmc->mmc_controls->boot_partitions) {
+            GList *input = emmc->mmc_controls->boot_partitions->input;
+            for (GList *b = input; b != NULL; b = b->next) {
+                PuEmmcBinary *bin = b->data;
+                g_free(bin->input->filename);
+                g_free(bin->input->md5sum);
+                g_free(bin->input->sha256sum);
+                g_free(bin->input);
+                g_free(bin);
+            }
+            g_list_free(g_steal_pointer(&input));
+            g_free(emmc->mmc_controls->boot_partitions);
+        }
+        g_free(emmc->mmc_controls);
     }
 
     if (emmc->disk)
@@ -536,60 +577,93 @@ pu_emmc_get_partition_table_size(PuEmmc *emmc)
 }
 
 static gboolean
-pu_emmc_parse_emmc_bootpart(PuEmmc *emmc,
-                            GHashTable *root,
-                            GError **error)
+pu_emmc_parse_mmc_controls(PuEmmc *emmc,
+                           GHashTable *root,
+                           GError **error)
 {
-    PuConfigValue *value_bootpart = g_hash_table_lookup(root, "emmc-boot-partitions");
+    PuConfigValue *value_mmc;
+    PuConfigValue *value_bootpart;
 
     g_return_val_if_fail(emmc != NULL, FALSE);
     g_return_val_if_fail(root != NULL, FALSE);
     g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-    if (!value_bootpart) {
-        g_debug("No entry 'emmc-boot-partitions' found. Skipping...");
+    value_mmc = g_hash_table_lookup(root, "mmc");
+    if (!value_mmc) {
+        g_debug("No entry 'mmc' found. Skipping...");
         return TRUE;
     }
+    if (value_mmc->type != PU_CONFIG_VALUE_TYPE_MAPPING) {
+        g_set_error(error, PU_ERROR, PU_ERROR_EMMC_PARSE,
+                    "'mmc' is not a valid YAML mapping");
+        return FALSE;
+    }
 
+    GHashTable *mmc_control = value_mmc->data.mapping;
+    emmc->mmc_controls = g_new0(PuEmmcControls, 1);
+    emmc->mmc_controls->hwreset=
+        pu_hash_table_lookup_string(mmc_control, "hwreset", NULL);
+    emmc->mmc_controls->bootbus =
+        pu_hash_table_lookup_string(mmc_control, "bootbus", NULL);
+
+    value_bootpart = g_hash_table_lookup(mmc_control, "boot-partitions");
+    if (!value_bootpart) {
+        g_debug("No entry 'boot-partitions' found. Skipping...");
+        return TRUE;
+    }
     if (value_bootpart->type != PU_CONFIG_VALUE_TYPE_MAPPING) {
         g_set_error(error, PU_ERROR, PU_ERROR_EMMC_PARSE,
-                    "'emmc-boot-partitions' is not a valid YAML mapping");
+                    "'boot-partitions' is not a valid YAML mapping");
         return FALSE;
     }
 
     GHashTable *bootpart = value_bootpart->data.mapping;
-    emmc->emmc_boot_partitions = g_new0(PuEmmcBootPartitions, 1);
-    emmc->emmc_boot_partitions->enable =
-        pu_hash_table_lookup_boolean(bootpart, "enable", FALSE);
-    emmc->emmc_boot_partitions->input_offset =
-        pu_hash_table_lookup_sector(bootpart, emmc->device, "input-offset", 0);
-    emmc->emmc_boot_partitions->output_offset =
-        pu_hash_table_lookup_sector(bootpart, emmc->device, "output-offset", 0);
+    emmc->mmc_controls->boot_partitions = g_new0(PuEmmcBootPartitions, 1);
+    emmc->mmc_controls->boot_partitions->enable =
+        pu_hash_table_lookup_int64(bootpart, "enable", 0);
+    emmc->mmc_controls->boot_partitions->boot_ack =
+        pu_hash_table_lookup_boolean(bootpart, "boot-ack", FALSE);
 
-    g_debug("Parsed bootpart: enable=%s input-offset=%lld output-offset=%lld",
-            emmc->emmc_boot_partitions->enable ? "true" : "false",
-            emmc->emmc_boot_partitions->input_offset,
-            emmc->emmc_boot_partitions->output_offset);
+    g_debug("Parsed bootpart: enable=%d boot-ack=%s",
+            emmc->mmc_controls->boot_partitions->enable,
+            emmc->mmc_controls->boot_partitions->boot_ack ? "true" : "false");
 
-    PuConfigValue *value_input = g_hash_table_lookup(bootpart, "input");
-    if (value_input == NULL) {
-        g_set_error(error, PU_ERROR, PU_ERROR_EMMC_PARSE,
-                    "No input specified for 'emmc-boot-partitions'");
-        return FALSE;
+    GList *input_list = pu_hash_table_lookup_list(bootpart, "binaries", NULL);
+    for (GList *b = input_list; b != NULL; b = b->next) {
+        PuConfigValue *v = b->data;
+        PuEmmcBinary *bin = g_new0(PuEmmcBinary, 1);
+        bin->input_offset = pu_hash_table_lookup_sector(v->data.mapping, emmc->device,
+                                                        "input-offset", 0);
+        bin->output_offset = pu_hash_table_lookup_sector(v->data.mapping, emmc->device,
+                                                         "output-offset", 0);
+
+        PuConfigValue *value_input = g_hash_table_lookup(v->data.mapping, "input");
+        if (value_input == NULL) {
+            g_set_error(error, PU_ERROR, PU_ERROR_EMMC_PARSE,
+                        "No input specified for boot partition binaries");
+            return FALSE;
+        }
+        if (value_input->type != PU_CONFIG_VALUE_TYPE_MAPPING) {
+            g_set_error(error, PU_ERROR, PU_ERROR_EMMC_PARSE,
+                        "'input' of binary does not contain a mapping");
+            return FALSE;
+        }
+
+        PuEmmcInput *input = g_new0(PuEmmcInput, 1);
+        input->filename = pu_hash_table_lookup_string(value_input->data.mapping, "filename", "");
+        input->md5sum = pu_hash_table_lookup_string(value_input->data.mapping, "md5sum", "");
+        input->sha256sum = pu_hash_table_lookup_string(value_input->data.mapping, "sha256sum", "");
+
+        bin->input = input;
+        g_debug("Parsed bootpart input: filename=%s md5sum=%s sha256sum=%s",
+                input->filename, input->md5sum, input->sha256sum);
+
+        emmc->mmc_controls->boot_partitions->input =
+            g_list_prepend(emmc->mmc_controls->boot_partitions->input, bin);
     }
-    if (value_input->type != PU_CONFIG_VALUE_TYPE_MAPPING) {
-        g_set_error(error, PU_ERROR, PU_ERROR_EMMC_PARSE,
-                    "'input' of binary does not contain a mapping");
-        return FALSE;
-    }
-    PuEmmcInput *input = g_new0(PuEmmcInput, 1);
-    input->uri = pu_hash_table_lookup_string(value_input->data.mapping, "uri", "");
-    input->md5sum = pu_hash_table_lookup_string(value_input->data.mapping, "md5sum", "");
-    input->sha256sum = pu_hash_table_lookup_string(value_input->data.mapping, "sha256sum", "");
-    emmc->emmc_boot_partitions->input = input;
 
-    g_debug("Parsed bootpart input: uri=%s md5sum=%s sha256sum=%s",
-            input->uri, input->md5sum, input->sha256sum);
+    emmc->mmc_controls->boot_partitions->input
+        = g_list_reverse(emmc->mmc_controls->boot_partitions->input);
 
     return TRUE;
 }
@@ -682,11 +756,11 @@ pu_emmc_parse_raw(PuEmmc *emmc,
             return FALSE;
         }
         PuEmmcInput *input = g_new0(PuEmmcInput, 1);
-        input->uri = pu_hash_table_lookup_string(value_input->data.mapping, "uri", "");
+        input->filename = pu_hash_table_lookup_string(value_input->data.mapping, "filename", "");
         input->md5sum = pu_hash_table_lookup_string(value_input->data.mapping, "md5sum", "");
         input->sha256sum = pu_hash_table_lookup_string(value_input->data.mapping, "sha256sum", "");
 
-        path = pu_path_from_uri(input->uri, prefix, error);
+        path = pu_path_from_filename(input->filename, prefix, error);
         if (path == NULL)
             return FALSE;
 
@@ -695,8 +769,8 @@ pu_emmc_parse_raw(PuEmmc *emmc,
             return FALSE;
 
         bin->input = input;
-        g_debug("Parsed raw input: uri=%s md5sum=%s sha256sum=%s",
-                input->uri, input->md5sum, input->sha256sum);
+        g_debug("Parsed raw input: filename=%s md5sum=%s sha256sum=%s",
+                input->filename, input->md5sum, input->sha256sum);
 
         PedSector min_offset = pu_emmc_get_partition_table_size(emmc);
 
@@ -752,9 +826,10 @@ pu_emmc_parse_partitions(PuEmmc *emmc,
         }
 
         PuEmmcPartition *part = g_new0(PuEmmcPartition, 1);
-        part->label = pu_hash_table_lookup_string(v->data.mapping, "label", "");
+        part->label = pu_hash_table_lookup_string(v->data.mapping, "label", NULL);
         part->partuuid = pu_hash_table_lookup_string(v->data.mapping, "partuuid", "");
         part->filesystem = pu_hash_table_lookup_string(v->data.mapping, "filesystem", "fat32");
+        part->mkfs_extra_args = pu_hash_table_lookup_string(v->data.mapping, "mkfs-extra-args", NULL);
         part->size = pu_hash_table_lookup_sector(v->data.mapping, emmc->device, "size", 0);
         part->offset = pu_hash_table_lookup_sector(v->data.mapping, emmc->device, "offset", 0);
         part->block_size = pu_hash_table_lookup_sector(v->data.mapping, emmc->device, "block-size", 2);
@@ -789,10 +864,11 @@ pu_emmc_parse_partitions(PuEmmc *emmc,
             }
         }
 
-        g_debug("Parsed partition: label=%s filesystem=%s type=%s size=%lld "
-                "offset=%lld block-size=%lld expand=%s",
-                part->label, part->filesystem, type_str, part->size, part->offset,
-                part->block_size, part->expand ? "true" : "false");
+        g_debug("Parsed partition: label=%s filesystem=%s mkfs-extra-args=%s "
+                "type=%s size=%lld offset=%lld block-size=%lld expand=%s",
+                part->label, part->filesystem, part->mkfs_extra_args, type_str,
+                part->size, part->offset, part->block_size,
+                part->expand ? "true" : "false");
 
         GList *flag_list = pu_hash_table_lookup_list(v->data.mapping, "flags", NULL);
         if (flag_list != NULL) {
@@ -826,13 +902,13 @@ pu_emmc_parse_partitions(PuEmmc *emmc,
                 }
 
                 PuEmmcInput *input = g_new0(PuEmmcInput, 1);
-                input->uri = pu_hash_table_lookup_string(iv->data.mapping, "uri", "");
+                input->filename = pu_hash_table_lookup_string(iv->data.mapping, "filename", "");
                 input->md5sum = pu_hash_table_lookup_string(iv->data.mapping, "md5sum", "");
                 input->sha256sum = pu_hash_table_lookup_string(iv->data.mapping, "sha256sum", "");
                 part->input = g_list_prepend(part->input, input);
 
-                g_debug("Parsed partition input: uri=%s md5sum=%s sha256sum=%s",
-                        input->uri, input->md5sum, input->sha256sum);
+                g_debug("Parsed partition input: filename=%s md5sum=%s sha256sum=%s",
+                        input->filename, input->md5sum, input->sha256sum);
             }
             part->input = g_list_reverse(part->input);
         }
@@ -963,7 +1039,7 @@ pu_emmc_new(const gchar *device_path,
         return NULL;
     }
 
-    if (!pu_emmc_parse_emmc_bootpart(self, root, error))
+    if (!pu_emmc_parse_mmc_controls(self, root, error))
         return NULL;
     if (!pu_emmc_parse_raw(self, root, error))
         return NULL;
